/*
 *    TTTTTTTTTTTTTT  EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *    TTTTTTTTTTTTTT  EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EEEEEEEEEE      OO          OO
 *          TT        EEEEEEEEEE      OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *          TT        EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *
 *                  L'mulateur Thomson TO8
 *
 *  Copyright (C) 1997-2001 Gilles Ftis, Eric Botcazou, Alexandre Pukall,
 *                          Jrmie Guillaume
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : dos/disk.c
 *  Version    : 1.7
 *  Cr par   : Alexandre Pukall mai 1998
 *  Modifi par: Eric Botcazou 20/09/2001
 *
 *  Lecture directe des disquettes Thomson.
 */


#ifndef SCAN_DEPEND
   #include <stdio.h>
   #include <stddef.h>
   #include <bios.h>
   #include <sys/movedata.h>
#endif

#include "to8.h"


#define NBSECT      16
#define DISK_RETRY   3

struct floppy_cmd {
    int cmd;
    int track;
    int sector;
    int nsects;
    void *buffer;
};

static int dpt_addr = 0;
static unsigned char pc_data[3];
static unsigned char format_filler_byte = 0;



/* SetDiskParameters:
 *  Initialise les paramtres disquette pour le format Thomson.
 */
static void SetDiskParameters(void)
{
    unsigned char to_data[3];

    /* on rcupre le vecteur 0x1E du BIOS qui pointe sur la table
       des paramtres de la disquette */
    _dosmemgetl(0x1E*4, 1, &dpt_addr);

    /* on sauvegarde les paramtres originaux */
    _dosmemgetb(dpt_addr+3, 3, pc_data);

    /* on fixe les nouveaux paramtres */
    to_data[0]=0x01;  /* 256 bytes/sector    */
    to_data[1]=0x10;  /* 16 sectors/track    */
    to_data[2]=0x1B;  /* gap between sectors */
    _dosmemputb(to_data, 3, dpt_addr+3);

    /* reset */
    biosdisk(0, 0, 0, 0, 1, 0, NULL);
}



/* SetFormatParameters:
 *  Initialise les paramtres spcifiques de formatage pour le format Thomson.
 */ 
static void SetFormatParameters(unsigned char filler_byte)
{
    unsigned char data[2];

    if (!dpt_addr)
        SetDiskParameters();

    data[0]=0x2C;  /* gap length when formatting */
    data[1]=filler_byte;
    _dosmemputb(data, 2, dpt_addr+7);
    biosdisk(0, 0, 0, 0, 1, 0, NULL);

    /* slection du format de formatage 720 ko */
    biosdisk(23, 0, 0, 0, 1, 4, NULL);

    format_filler_byte = filler_byte;
}



/* ExecCommand:
 *  Excute la commande spcifie (via l'interruption 13h du BIOS).
 */
static int ExecCommand(struct floppy_cmd *fd_cmd)
{
    int i;
    int ret=0x10;

    if (!dpt_addr)
        SetDiskParameters();

    for (i=0; i<DISK_RETRY; i++)
    {
        ret=biosdisk(fd_cmd->cmd, 0, 0, fd_cmd->track, fd_cmd->sector,
                                                       fd_cmd->nsects, fd_cmd->buffer);
        /* commande OK */
        if (ret==0)
            break;

        /* commande non OK mais corrige par ctrl */
        if ((i>1) && (ret==0x11))
        {
            ret=0;
            break;
        }

        /* reset du lecteur */
        biosdisk(0, 0, 0, 0, 1, 0, NULL);
    }

    switch (ret)
    {
        case 0x02:  /* address mark not found */
            return 0x04;   /* erreur sur l'adresse */

        case 0x03:  /* disk write-protected */
            return 0x01;   /* disk protg en criture */

        case 0x04:  /* sector not found */
        case 0x07:  /* drive parameter activity failed */
        case 0x10:  /* data read (CRC or ECC) error */
        case 0x0A:  /* bad sector flag detected */
        case 0x0B:  /* bad track flag detected */
            return 0x08;   /* erreur sur les donnes */

        case 0x06:  /* floppy disk removed */   
        case 0x80:  /* disk timed out or failed to respond */
            return 0x10;   /* lecteur non prt */

        default:
            return 0;  /* OK */
    }
}



/* ReadSector:
 *  Lit le secteur spcifi sur la disquette.
 */
static int ReadSector(int drive, int track, int sector, int nsects, unsigned char data[])
{
    struct floppy_cmd fd_cmd;

    fd_cmd.cmd    = 2;  
    fd_cmd.track  = track;
    fd_cmd.sector = sector;
    fd_cmd.nsects = nsects;
    fd_cmd.buffer = data;

    return ExecCommand(&fd_cmd);
    (void) drive;
}



/* WriteSector:
 *  Ecrit le secteur spcifi sur la disquette.
 */
static int WriteSector(int drive, int track, int sector, int nsects, unsigned char data[])
{
    struct floppy_cmd fd_cmd;

    fd_cmd.cmd    = 3;  
    fd_cmd.track  = track;
    fd_cmd.sector = sector;
    fd_cmd.nsects = nsects;
    fd_cmd.buffer = data;

    return ExecCommand(&fd_cmd);
    (void) drive;
}



/* FormatTrack:
 *  Formate la piste en utilisant la table des headers spcifie.
 */
static int FormatTrack(int drive, int track, unsigned char headers_table[], 
                                                      unsigned char filler_byte)
{
    struct floppy_cmd fd_cmd;

    if (filler_byte != format_filler_byte)
        SetFormatParameters(filler_byte);
     
    fd_cmd.cmd    = 5;  
    fd_cmd.track  = track;
    fd_cmd.sector = 1;
    fd_cmd.nsects = NBSECT;
    fd_cmd.buffer = headers_table;

    return ExecCommand(&fd_cmd);
    (void) drive;
}



/* InitDirectDisk:
 *  Initialise le module de lecture directe.
 */
void InitDirectDisk(int enable_write)
{
    to8_DirectReadSector = ReadSector;

    if (enable_write)
    {
        to8_DirectWriteSector = WriteSector;
        to8_DirectFormatTrack = FormatTrack;
    }
}



/* ExitDirectDisk:
 *  Met au repos le module de lecture directe.
 */
void ExitDirectDisk(void)
{
    to8_DirectReadSector = NULL;
    to8_DirectWriteSector = NULL;
    to8_DirectFormatTrack = NULL;

    if (dpt_addr)
    {
        /* restaure les paramtres originaux */
        _dosmemputb(pc_data, 3, dpt_addr+3);

        /* reset */
        biosdisk(0, 0, 0, 0, 1, 0, NULL);

        dpt_addr = 0;
    }
}

