/*
 *    TTTTTTTTTTTTTT  EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *    TTTTTTTTTTTTTT  EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EEEEEEEEEE      OO          OO
 *          TT        EEEEEEEEEE      OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EE              OO          OO
 *          TT        EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *          TT        EEEEEEEEEEEEEE  OOOOOOOOOOOOOO
 *
 *                  L'mulateur Thomson TO8
 *
 *  Copyright (C) 1997-2001 Gilles Ftis, Eric Botcazou, Alexandre Pukall,
 *                          Jrmie Guillaume
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : linux/disk.c
 *  Version    : 1.7
 *  Cr par   : Eric Botcazou 29/07/2000
 *  Modifi par: Eric Botcazou 26/09/2001
 *
 *  Lecture directe des disquettes Thomson.
 */


#ifndef SCAN_DEPEND
   #include <stdio.h>
   #include <fcntl.h>
   #include <unistd.h>
   #include <sys/ioctl.h>
   #include <linux/fd.h>
   #include <linux/fdreg.h>
#endif

#include "to8.h"


#define NBSECT      16
#define DISK_RETRY   3

static int fd = -1;



/* ResetFloppyDrive:
 *  Rinitialise le lecteur de disquettes.
 */
static void ResetFloppyDrive(void)
{
    static struct floppy_struct fd_prm={ 1280,  /* size    */
					 8,     /* sect    */
					 2,     /* head    */
					 80,    /* track   */
					 0,     /* stretch */
					 0x1B,  /* gap     */
					 0x3A,  /* rate    */
					 0xDF,  /* spec1   */
					 0x2C,  /* fmt_gap */ };
    ioctl(fd, FDSETPRM, &fd_prm);
}



/* OpenFloppyDrive:
 *  Obtient le descripteur de fichier pour le lecteur de disquettes.
 */
static int OpenFloppyDrive(void)
{
    if (to8_DirectWriteSector)
    {
	if ( (fd=open("/dev/fd0", O_RDWR | O_NDELAY))<0 )
	{
	    to8_DirectWriteSector = NULL;
	    to8_DirectFormatTrack = NULL;
	}
        else
            goto success;
    }

    if ( (fd=open("/dev/fd0", O_RDONLY | O_NDELAY))<0 )
	to8_DirectReadSector = NULL;
    else
	goto success;

    return 0;

  success:
     ResetFloppyDrive();
     return 1;
}



/* ExecCommand:
 *  Excute la commande spcifie via l'appel ioctl() FDRAWCMD.
 */
static int ExecCommand(struct floppy_raw_cmd *fd_cmd)
{
    int i, ret;

    if (fd<0 && !OpenFloppyDrive())
	return 0x10;  /* lecteur non prt */

    for (i=0; i<DISK_RETRY; i++)
    {
        ret=ioctl(fd, FDRAWCMD, fd_cmd);

	if (ret >= 0)
	    break;

	ResetFloppyDrive();
    }

    if (ret<0)
	return 0x10;  /* lecteur non prt */

#ifdef DEBUG
    printf("fd_cmd reply: ");
    for (i=0; i<3; i++)
	printf("ST%d=%02x ", i, fd_cmd->reply[i]);
    printf("\n");
#endif

    switch (fd_cmd->reply[1])  /* ST1 */
    {
        case 0x01:  /* Missing Address Mark */
	    return 0x04;   /* erreur sur l'adresse */

	case 0x02:  /* Write Protected */
	    return 0x01;   /* disk protg en criture */

        case 0x04:  /* No Data - unreadable */
	    return 0x08;   /* erreur sur les donnes */

	case 0x20:  /* CRC error in data or addr */
	    if (fd_cmd->reply[2]==0x20)
		return 0x08;   /* erreur sur les donnes */
	    else
		return 0x04;   /* erreur sur l'adresse */

        default:
	    return 0;  /* OK */
    }
}



/* ReadSector:
 *  Lit le secteur spcifi sur la disquette.
 */
static int ReadSector(int drive, int track, int sector, int nsects, unsigned char data[])
{
    struct floppy_raw_cmd fd_cmd;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_READ | FD_RAW_NEED_SEEK;
    fd_cmd.data   = data;
    fd_cmd.length = 256*nsects; /* SECTOR_SIZE */
    fd_cmd.rate   = 2;
    fd_cmd.track  = track;

    fd_cmd.cmd[0] = FD_READ;
    fd_cmd.cmd[1] = drive << 2;
    fd_cmd.cmd[2] = track;
    fd_cmd.cmd[3] = 0;  /* head */
    fd_cmd.cmd[4] = sector;
    fd_cmd.cmd[5] = 1;  /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[6] = NBSECT;
    fd_cmd.cmd[7] = 0x1B;
    fd_cmd.cmd[8] = 0xFF;
    fd_cmd.cmd_count = 9;

    return ExecCommand(&fd_cmd);
}



/* WriteSector:
 *  Ecrit le secteur spcifi sur la disquette.
 */
static int WriteSector(int drive, int track, int sector, int nsects,unsigned char data[])
{
    struct floppy_raw_cmd fd_cmd;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_WRITE | FD_RAW_NEED_SEEK;
    fd_cmd.data   = data;
    fd_cmd.length = 256*nsects; /* SECTOR_SIZE */
    fd_cmd.rate   = 2;
    fd_cmd.track  = track;

    fd_cmd.cmd[0] = FD_WRITE;
    fd_cmd.cmd[1] = drive << 2;
    fd_cmd.cmd[2] = track;
    fd_cmd.cmd[3] = 0;  /* head */
    fd_cmd.cmd[4] = sector;
    fd_cmd.cmd[5] = 1;  /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[6] = NBSECT;
    fd_cmd.cmd[7] = 0x1B;
    fd_cmd.cmd[8] = 0xFF;
    fd_cmd.cmd_count = 9;

    return ExecCommand(&fd_cmd);
}



/* FormatTrack:
 *  Formate la piste en utilisant la table des headers spcifie.
 */
static int FormatTrack(int drive, int track, unsigned char headers_table[],
                                                     unsigned char filler_byte)
{
    struct floppy_raw_cmd fd_cmd;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_WRITE | FD_RAW_NEED_SEEK;
    fd_cmd.data   = headers_table;
    fd_cmd.length = 64;
    fd_cmd.rate   = 2;
    fd_cmd.track  = track;

    fd_cmd.cmd[0] = FD_FORMAT;
    fd_cmd.cmd[1] = drive << 2;
    fd_cmd.cmd[2] = 1;  /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[3] = NBSECT;
    fd_cmd.cmd[4] = 0x2C;  /* fmt_gap */
    fd_cmd.cmd[5] = filler_byte;
    fd_cmd.cmd_count = 6;

    return ExecCommand(&fd_cmd);
}



/* InitDirectDisk:
 *  Initialise le module de lecture directe.
 */
void InitDirectDisk(int enable_write)
{
    to8_DirectReadSector = ReadSector;

    if (enable_write)
    {
        to8_DirectWriteSector = WriteSector;
        to8_DirectFormatTrack = FormatTrack;
    }
}



/* ExitDirectDisk:
 *  Met au repos le module de lecture directe.
 */
void ExitDirectDisk(void)
{
    to8_DirectReadSector = NULL;
    to8_DirectWriteSector = NULL;
    to8_DirectFormatTrack = NULL;

    if (fd >= 0)
    {
        close(fd);
        fd = -1;
    }
}
